Imports System.Text
Imports System.IO
Imports System.Net
Imports System.Threading
Imports System.Collections.Concurrent
Imports System.Drawing.Drawing2D
Imports System.Threading.Tasks

Namespace FileDownloader
	Partial Public Class Form1
		Inherits Form

		'0 = not processed
		'1 = processed and success
		'2 = processed and failure
		Private _tasks As New ConcurrentDictionary(Of Tuple(Of Long, Long), Integer)

		Private _progress As New ProgressPanel
		Private _url As String = String.Empty
		Private _fileName As String = String.Empty
		Private _authorization As String = String.Empty

		Private _cancellationTokenSource As CancellationTokenSource = CancellationTokenSource.CreateLinkedTokenSource(New CancellationToken(False))

		Public Sub New()
			InitializeComponent()

			Me.lblCPUCount.Text = Environment.ProcessorCount.ToString
			Me.numUpDownThreads.Value = Environment.ProcessorCount

			_progress.Spring = True
			_progress.BorderSides = ToolStripStatusLabelBorderSides.All
			_progress.BorderStyle = Border3DStyle.SunkenOuter
			_progress.Overflow = ToolStripItemOverflow.Never
			Me.statusStrip1.AutoSize = False
			Me.statusStrip1.Height = 30
			Me.statusStrip1.Items.Add(CType(_progress, ToolStripItem))

			Dim byteStatus As New ToolStripStatusLabel("0 of 0")
			byteStatus.Name = "byteStatus"
			Me.statusStrip1.Items.Add(byteStatus)
		End Sub

		Private Sub Form1_Load(ByVal sender As Object, ByVal e As EventArgs) Handles MyBase.Load
			ServicePointManager.DefaultConnectionLimit = 500
			ServicePointManager.SecurityProtocol = SecurityProtocolType.Ssl3

'			#Region "Use to Debug SSL Certificate Problems"
			'ServicePointManager.ServerCertificateValidationCallback = (object request, System.Security.Cryptography.X509Certificates.X509Certificate certificate, System.Security.Cryptography.X509Certificates.X509Chain chain, System.Net.Security.SslPolicyErrors sslPolicyErrors) =>
			'{
			'    try
			'    {
			'        if (sslPolicyErrors == System.Net.Security.SslPolicyErrors.None)
			'        {
			'            return true;
			'        }
			'        else if (sslPolicyErrors == System.Net.Security.SslPolicyErrors.RemoteCertificateNameMismatch)
			'        {
			'            HttpWebRequest webRequest = (HttpWebRequest)request;

			'            System.Security.Policy.Zone zone = System.Security.Policy.Zone.CreateFromUrl(webRequest.RequestUri.ToString);
			'            if (zone.SecurityZone == System.Security.SecurityZone.MyComputer)  // || zone.SecurityZone == System.Security.SecurityZone.Intranet
			'                return true;

			'            string name = certificate.Issuer;
			'            name = name.Replace("CN=", "");
			'            if (name.ToLower == Environment.MachineName.ToLower)
			'                return true;
			'        }
			'        return false;
			'    }
			'    catch
			'    {
			'        return false;
			'    }
			'};
'			#End Region
		End Sub

		Private Sub Form1_FormClosing(ByVal sender As Object, ByVal e As FormClosingEventArgs) Handles MyBase.FormClosing
		End Sub

		Private Sub Form1_Shown(ByVal sender As Object, ByVal e As EventArgs) Handles MyBase.Shown
		End Sub

		#Region "Download"
		Private Sub btnDownload_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnDownload.Click
			Try
				btnRetry.Enabled = False
				btnDownload.Enabled = False

				If Me.grpBoxAuthentication.Enabled Then
					If Me.txtUser.Text.Trim.Length > 0 AndAlso Me.txtPassword.Text.Trim.Length > 0 Then
						_authorization = Convert.ToBase64String(Encoding.ASCII.GetBytes(txtUser.Text & ":" & txtPassword.Text))
					End If
				End If

				SetFormState(False)

				lblResult.Visible = False
				lblResult.ForeColor = Color.Blue
				lblResult.Text = ""

				_tasks.Clear()

				_progress.ClearSegments()
				Me.statusStrip1.Items("byteStatus").Text = "0 of 0"
				Me.statusStrip1.Refresh()

				_url = txtURL.Text
				_fileName = GetFileNameFromQueryString(New Uri(_url))
				If String.IsNullOrEmpty(_fileName) Then
					Throw New Exception("'?file=yourfile.ext' was not specified in request query string.")
				End If

				btnCancel.Enabled = chkSplitFileIntoChunks.Checked

				StartProgressIndicator()

				DownloadFile(txtURL.Text)
			Catch ex As Exception
				StopProgressIndicator()

				MessageBox.Show("Download Error:  " & ex.GetBaseException.Message)
				btnRetry.Enabled = False
				btnDownload.Enabled = True
				btnCancel.Enabled = False

				SetFormState(True)

				_authorization = String.Empty
			End Try
		End Sub

		Private Sub DownloadFile(ByVal url As String)
			'get chunk size
			Dim chunkSize As Long = 0
			If chkSplitFileIntoChunks.Checked Then
				chunkSize = CLng(Fix(numUpDownChunkSize.Value))
			End If

			'get file length
			Dim fileLength As Long = 0
			Dim webRequest As HttpWebRequest = CType(HttpWebRequest.Create(url), HttpWebRequest)
			webRequest.Method = "HEAD"
			If Not String.IsNullOrEmpty(_authorization) Then
				webRequest.Headers.Add("authorization", "Basic " & _authorization)
			End If
			Using webResponse As HttpWebResponse = CType(webRequest.GetResponse, HttpWebResponse)
				fileLength = webResponse.ContentLength
			End Using
			_progress.TotalFileLength = fileLength
			Me.statusStrip1.Items("byteStatus").Text = "0 of " & fileLength.ToString
			Me.statusStrip1.Refresh()

			Using fs As New FileStream(_fileName, FileMode.Create, FileAccess.Write, FileShare.Write)
				fs.SetLength(fileLength)
			End Using

			'transfer entire file
			If chunkSize = 0 Then
				Dim bytesRange As New Tuple(Of Long,Long)(0, fileLength)
				If Not _tasks.ContainsKey(bytesRange) Then
					_tasks.TryAdd(bytesRange, 0)
				Else
					_tasks(bytesRange) = 0
				End If

				ThreadPool.QueueUserWorkItem(Sub(o) StartProcessEntireFile(o), New Tuple(Of Long, Long)(0, fileLength))
			Else 'transfer file in chunks
				ThreadPool.QueueUserWorkItem(Sub(o) StartProcessChunks(o), New Object() { fileLength, chunkSize })
			End If
		End Sub

		Private Sub StartProcessChunks(ByVal state As Object)
			Try
				Dim args() As Object = CType(state, Object)
				Dim options As New ParallelOptions
				options.MaxDegreeOfParallelism = CInt(Fix(numUpDownThreads.Value))
				options.CancellationToken = _cancellationTokenSource.Token
				Parallel.ForEach(SplitFileIntoChunks(CLng(Fix(args(0))), CLng(Fix(args(1)))), options, New Action(Of Tuple(Of Long, Long))(AddressOf ProcessChunks))

				Finished()
			Catch e1 As OperationCanceledException
				Canceled()
			End Try
		End Sub

		Private Function SplitFileIntoChunks(ByVal fileLength As Long, ByVal chunkSize As Long) As IEnumerable(Of Tuple(Of Long, Long))
            Dim index As Long = 0
            Dim returnTuple As New List(Of Tuple(Of Long, Long))

            Do
                If index >= fileLength Then
                    Exit Do
                End If
                If index < fileLength AndAlso index + chunkSize > fileLength Then
                    Dim bytesRange As New Tuple(Of Long, Long)(index, fileLength)
                    If Not _tasks.ContainsKey(bytesRange) Then
                        _tasks.TryAdd(bytesRange, 0)
                    Else
                        _tasks(bytesRange) = 0
                    End If
                    Dim ret As New Tuple(Of Long, Long)(index, fileLength)
                    returnTuple.Add(ret)

                    Exit Do
                End If
                Dim bytesRange2 As New Tuple(Of Long, Long)(index, (index + chunkSize))
                If Not _tasks.ContainsKey(bytesRange2) Then
                    _tasks.TryAdd(bytesRange2, 0)
                Else
                    _tasks(bytesRange2) = 0
                End If
                Dim ret2 As New Tuple(Of Long, Long)(index, (index + chunkSize))
                returnTuple.Add(ret2)
                index += chunkSize
            Loop
            Return returnTuple
        End Function

		Private Sub ProcessChunks(ByVal bytesRange As Tuple(Of Long, Long))
			Try
				Dim webRequest As HttpWebRequest = CType(HttpWebRequest.Create(_url), HttpWebRequest)
				webRequest.AddRange(bytesRange.Item1, bytesRange.Item2)
				If Not String.IsNullOrEmpty(_authorization) Then
					webRequest.Headers.Add("authorization", "Basic " & _authorization)
				End If

				Using webResponse As HttpWebResponse = CType(webRequest.GetResponse, HttpWebResponse)
					Dim range As String = webResponse.GetResponseHeader("content-range")
					Dim parts() As String = range.Split(New Char() { " "c, "-"c, "/"c })
					Dim contentLength As Long = Long.Parse(parts(2)) - Long.Parse(parts(1))
					Dim totalLength As Long = Long.Parse(parts(3))

					'const FileOptions FILE_NO_BUFFER = (FileOptions)0x20000000;
					'using (FileStream fs = new FileStream(_fileName, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite, 4096, FileOptions.WriteThrough | FILE_NO_BUFFER | FileOptions.Asynchronous))
					Using fs As New FileStream(_fileName, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite, 4096, True)
						fs.Position = Long.Parse(parts(1))
						Using s As Stream = webResponse.GetResponseStream
							s.CopyTo(fs)
						End Using
					End Using
				End Using

				'processing successful
				_tasks(bytesRange) = 1
				ChunkProcessed(bytesRange)
			Catch
				'processing failed
				_tasks(bytesRange) = 2
				ChunkFailed(bytesRange)
			End Try
		End Sub

		Private Sub StartProcessEntireFile(ByVal state As Object)
			Dim bytesRange As Tuple(Of Long, Long) = CType(state, Tuple(Of Long, Long))
			Try
				Dim webRequest As HttpWebRequest = CType(HttpWebRequest.Create(_url), HttpWebRequest)
				If Not String.IsNullOrEmpty(_authorization) Then
					webRequest.Headers.Add("authorization", "Basic " & _authorization)
				End If

				Using webResponse As HttpWebResponse = CType(webRequest.GetResponse, HttpWebResponse)
					Using fs As New FileStream(_fileName, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite, 4096, True)
						fs.Position = 0
						Using s As Stream = webResponse.GetResponseStream
							s.CopyTo(fs)
						End Using
					End Using
				End Using

				'processing successful
				_tasks(bytesRange) = 1
				ChunkProcessed(bytesRange)
			Catch
				'processing failed
				_tasks(bytesRange) = 2
				ChunkFailed(bytesRange)
			Finally
				Finished()
			End Try
		End Sub
		#End Region

		#Region "Retry"
		Private Sub btnRetry_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnRetry.Click
			Try
				btnDownload.Enabled = False
				btnRetry.Enabled = False
				btnCancel.Enabled = True

				If Me.grpBoxAuthentication.Enabled Then
					If Me.txtUser.Text.Trim.Length > 0 AndAlso Me.txtPassword.Text.Trim.Length > 0 Then
						_authorization = Convert.ToBase64String(Encoding.ASCII.GetBytes(txtUser.Text & ":" & txtPassword.Text))
					End If
				End If

				SetFormState(False)

				lblResult.Visible = False
				lblResult.ForeColor = Color.Blue
				lblResult.Text = ""

				_url = txtURL.Text
				_fileName = GetFileNameFromQueryString(New Uri(_url))
				If String.IsNullOrEmpty(_fileName) Then
					Throw New Exception("'?file=yourfile.ext' was not specified in request query string.")
				End If

				btnCancel.Enabled = chkSplitFileIntoChunks.Checked

				StartProgressIndicator()

				DownloadFileRetry(txtURL.Text)
			Catch ex As Exception
				StopProgressIndicator()

				MessageBox.Show("Retry Error:  " & ex.GetBaseException.Message)
				btnRetry.Enabled = True
				btnDownload.Enabled = True
				btnCancel.Enabled = False

				SetFormState(True)

				_authorization = String.Empty
			End Try
		End Sub

		Private Sub DownloadFileRetry(ByVal url As String)
			Dim chunkSize As Long = 0
			If chkSplitFileIntoChunks.Checked Then
				chunkSize = CLng(Fix(numUpDownChunkSize.Value))
			End If

            If chunkSize = 0 Then

                ''Dim fileLength As Long = CLng(Fix(_tasks.Keys.Sum(Of Tuple(Of Long, Long))(Function(e) (e.Item2 - e.Item1))))

                Dim fileLength As Long = CLng(Fix(_tasks.Keys.Sum(Function(e) (e.Item2 - e.Item1))))

                ThreadPool.QueueUserWorkItem(Sub(o) StartProcessEntireFile(o), New Tuple(Of Long, Long)(0, fileLength))
            Else
                ThreadPool.QueueUserWorkItem(Sub(o) StartProcessRetry(o))
			End If
		End Sub

		Private Sub StartProcessRetry(ByVal state As Object)
			Try
				Dim options As New ParallelOptions
				options.MaxDegreeOfParallelism = CInt(Fix(numUpDownThreads.Value))
				Parallel.ForEach(FindRetryChunks, options, New Action(Of Tuple(Of Long, Long))(AddressOf ProcessChunks))

				Finished()
			Catch e1 As OperationCanceledException
				Canceled()
			End Try
		End Sub

		Private Function FindRetryChunks() As IEnumerable(Of Tuple(Of Long, Long))
            Dim returnVal As New List(Of Tuple(Of Long, Long))
            For Each bytesRange As Tuple(Of Long, Long) In _tasks.Keys
                If _tasks(bytesRange) = 2 Then 'processed but failed
                    If Not _tasks.ContainsKey(bytesRange) Then
                        _tasks.TryAdd(bytesRange, 0)
                    Else
                        _tasks(bytesRange) = 0
                    End If
                    returnVal.Add(bytesRange)

                End If
            Next bytesRange
            Return returnVal

        End Function
#End Region

#Region "Cancel"
        Private Sub btnCancel_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnCancel.Click
            If _cancellationTokenSource IsNot Nothing AndAlso _cancellationTokenSource.Token.ToString IsNot Nothing AndAlso _cancellationTokenSource.Token.CanBeCanceled Then
                _cancellationTokenSource.Cancel()

                StartCancelationIndicator()
            End If
        End Sub
		#End Region

		Private Sub ChunkProcessed(ByVal bytesRangeOfChunk As Tuple(Of Long, Long))
			If Me.IsHandleCreated AndAlso (Not Me.IsDisposed) Then
				If Me.InvokeRequired Then
					Me.BeginInvoke(CType(Sub() ChunkProcessed(bytesRangeOfChunk), Action))
				Else
					_progress.AddSegment(bytesRangeOfChunk, True)
					Me.statusStrip1.Items("byteStatus").Text = _progress.GetSumSegments.ToString & " of " & _progress.TotalFileLength.ToString
					Me.statusStrip1.Refresh()
				End If
			End If
		End Sub

		Private Sub ChunkFailed(ByVal bytesRangeOfChunk As Tuple(Of Long, Long))
			If Me.IsHandleCreated AndAlso (Not Me.IsDisposed) Then
				If Me.InvokeRequired Then
					Me.BeginInvoke(CType(Sub() ChunkFailed(bytesRangeOfChunk), Action))
				Else
					_progress.UpdateSegment(bytesRangeOfChunk, False)
					Me.statusStrip1.Items("byteStatus").Text = _progress.GetSumSegments.ToString & " of " & _progress.TotalFileLength.ToString
					Me.statusStrip1.Refresh()
				End If
			End If
		End Sub

		Private Sub Finished()
			StopProgressIndicator()

			_authorization = String.Empty

			If Me.IsHandleCreated AndAlso (Not Me.IsDisposed) Then
				If Me.InvokeRequired Then
					Me.BeginInvoke(CType(Sub() Finished(), Action))
				Else
					If _progress.GetSumSegments = _progress.TotalFileLength Then
						lblResult.Visible = True
						lblResult.ForeColor = Color.Blue
						lblResult.Text = "Success!"

						btnRetry.Enabled = False
						btnDownload.Enabled = True
						btnCancel.Enabled = False

						SetFormState(True)
					Else
						lblResult.Visible = True
						lblResult.ForeColor = Color.Red
						lblResult.Text = "Failed!"

						btnRetry.Enabled = True
						btnDownload.Enabled = True
						btnCancel.Enabled = False

						SetFormState(True)
					End If
				End If
			End If
		End Sub

		Private Sub Canceled()
			StopProgressIndicator()

			StopCancelationIndicator()

			_authorization = String.Empty

			If Me.IsHandleCreated AndAlso (Not Me.IsDisposed) Then
				If Me.InvokeRequired Then
					Me.BeginInvoke(CType(Sub() Canceled(), Action))
				Else
					btnRetry.Enabled = False
					btnDownload.Enabled = True
					btnCancel.Enabled = False

					SetFormState(True)

					lblResult.Visible = True
					lblResult.ForeColor = Color.Red
					lblResult.Text = "Canceled!"

					_tasks.Clear()

					_progress.ClearSegments()
					Me.statusStrip1.Items("byteStatus").Text = "0 of 0"
					Me.statusStrip1.Refresh()

					_cancellationTokenSource.Dispose()
					_cancellationTokenSource = Nothing
					_cancellationTokenSource = CancellationTokenSource.CreateLinkedTokenSource(New CancellationToken(False))
				End If
			End If
		End Sub

		Private Function GetFileNameFromQueryString(ByVal uri As Uri) As String
			Dim fileName As String = String.Empty
			Dim query As String = uri.Query
			query = query.Replace("?", "")
			Dim queryParts() As String = query.Split("&"c)
			For Each queryPart As String In queryParts
				If queryPart.StartsWith("file=") Then
					Dim querySubParts() As String = queryPart.Split("="c, "\"c, "/"c)
					fileName = querySubParts(querySubParts.GetLength(0) - 1)
					Exit For
				End If
			Next queryPart
			fileName = fileName.Replace("%20", "_")
			Return fileName
		End Function

		Private Sub SetFormState(ByVal enabled As Boolean)
			If enabled Then
				chkSplitFileIntoChunks.Enabled = enabled
				chkBasicAuthorization.Enabled = enabled
				grpBoxChunks.Enabled = chkSplitFileIntoChunks.Checked
				grpBoxAuthentication.Enabled = chkBasicAuthorization.Checked
			Else
				chkSplitFileIntoChunks.Enabled = enabled
				chkBasicAuthorization.Enabled = enabled
				grpBoxChunks.Enabled = enabled
				grpBoxAuthentication.Enabled = enabled
			End If
		End Sub

		Private Sub txtURL_TextChanged(ByVal sender As Object, ByVal e As EventArgs) Handles txtURL.TextChanged
			btnDownload.Enabled = txtURL.Text.Trim.Length > 0
		End Sub

		Private Sub chkChunkSize_CheckedChanged(ByVal sender As Object, ByVal e As EventArgs) Handles chkSplitFileIntoChunks.CheckedChanged
			grpBoxChunks.Enabled = chkSplitFileIntoChunks.Checked
		End Sub

		Private Sub chkBasicAuthorization_CheckedChanged(ByVal sender As Object, ByVal e As EventArgs) Handles chkBasicAuthorization.CheckedChanged
			grpBoxAuthentication.Enabled = chkBasicAuthorization.Checked
		End Sub

		#Region "ProgressPanel"
		Private Class ProgressPanel
			Inherits ToolStripStatusLabel

			Private _segments As New Dictionary(Of Tuple(Of Long, Long), Boolean)
			Private _totalLength As Long = 0
			Private _text As String = String.Empty

			Public Sub ClearSegments()
				_segments.Clear()
			End Sub

			Public Sub AddSegment(ByVal segment As Tuple(Of Long, Long), ByVal success As Boolean)
				If Not _segments.ContainsKey(segment) Then
					_segments.Add(segment, success)
				Else
					_segments(segment) = success
				End If
			End Sub

			Public Sub UpdateSegment(ByVal segment As Tuple(Of Long, Long), ByVal success As Boolean)
				If Not _segments.ContainsKey(segment) Then
					_segments.Add(segment, success)
				Else
					_segments(segment) = success
				End If
			End Sub

			Public Sub RemoveSegment(ByVal segment As Tuple(Of Long, Long))
				_segments.Remove(segment)
			End Sub

			Public Overloads Property Text As String
				Get
					Return _text
				End Get
				Set(ByVal value As String)
					_text = value
				End Set
			End Property

			Public Property TotalFileLength As Long
				Get
					Return _totalLength
				End Get
				Set(ByVal value As Long)
					_totalLength = value
				End Set
			End Property

			Public Function GetSumSegments() As Long
				Dim sum As Long = 0
				For Each key As Tuple(Of Long, Long) In _segments.Keys
					If _segments(key) = True Then 'only sum successful
						sum += key.Item2 - key.Item1
					End If
				Next key
				Return sum
			End Function

			Protected Overrides Sub OnPaint(ByVal e As PaintEventArgs)
				e.Graphics.TranslateTransform(0f, 1f)

				MyBase.OnPaint(e)

				For Each segment As Tuple(Of Long, Long) In _segments.Keys
					Dim begin As Long = segment.Item1
					Dim [end] As Long = segment.Item2

					Dim beginPercent As Double = CDbl(begin) / CDbl(_totalLength)
					Dim endPercent As Double = CDbl([end]) / CDbl(_totalLength)

					Dim left As Double = (CDbl(Me.Width) - 4R) * beginPercent
					Dim right As Double = (CDbl(Me.Width) - 4R) * endPercent

					Dim rect As New RectangleF(CSng(left) + 1f, 1f, CSng(right) - CSng(left), CSng(Me.Height) - 4f)

					If _segments(segment) = True Then
						Using b As New LinearGradientBrush(rect, Color.Blue, Color.White, LinearGradientMode.Vertical)
							b.SetBlendTriangularShape(0.5f, 0.65f)
							e.Graphics.FillRectangle(b, rect)
						End Using
					Else
						e.Graphics.FillRectangle(Brushes.Red, rect)

						Using pen As New Pen(Color.FromArgb(30, Color.Black), 3f)
							Dim mode As SmoothingMode = e.Graphics.SmoothingMode
							Try
								e.Graphics.SmoothingMode = SmoothingMode.AntiAlias
								e.Graphics.DrawLine(pen, New PointF(rect.Left, rect.Top), New PointF(rect.Right, rect.Bottom))
								e.Graphics.DrawLine(pen, New PointF(rect.Left, rect.Bottom), New PointF(rect.Right, rect.Top))
							Finally
								e.Graphics.SmoothingMode = mode
							End Try
						End Using
					End If

					Using pen As New Pen(Color.White)
						e.Graphics.DrawLine(pen, New PointF(rect.Left, rect.Top), New PointF(rect.Left, rect.Bottom))
						e.Graphics.DrawLine(pen, New PointF(rect.Right, rect.Top), New PointF(rect.Right, rect.Bottom))
					End Using
				Next segment

				If Not String.IsNullOrEmpty(_text) Then
					Dim size As SizeF = e.Graphics.MeasureString(_text, Me.Font)
					Dim x As Single = (CSng(Me.Width) / 2.0f) - (size.Width / 2.0f)
					Dim y As Single = (CSng(Me.Height) / 2.0f) - (size.Height / 2.0f)
					Dim pt As New PointF(x, y)
					Using b As Brush = New SolidBrush(Color.Red)
						e.Graphics.DrawString(_text, Me.Font, b, pt)
					End Using
				End If
			End Sub
		End Class
		#End Region

		#Region "ProgressIndicator"
		Private _cancelProgressIndicator As Boolean = True

		Private Sub StopProgressIndicator()
			_cancelProgressIndicator = True
		End Sub

		Private Sub StartProgressIndicator()
			_cancelProgressIndicator = False
			ThreadPool.QueueUserWorkItem(Sub(o) AnimateProgressIndicator(o))
		End Sub

		Private Sub AnimateProgressIndicator(ByVal state As Object)
			Dim toggle? As Boolean = True
			Do While Not _cancelProgressIndicator
				UpdateProgressIndicator(toggle)
				Thread.Sleep(500)
				toggle = Not toggle
			Loop
			UpdateProgressIndicator(Nothing)
		End Sub

		Private Sub UpdateProgressIndicator(ByVal toggle? As Boolean)
			If Me.IsHandleCreated AndAlso (Not Me.IsDisposed) Then
				If Me.InvokeRequired Then
					Me.BeginInvoke(CType(Sub() UpdateProgressIndicator(toggle), Action))
				Else
					If toggle.HasValue Then
						If toggle.Value Then
							Me.Text = "File Downloader  \"
						Else
							Me.Text = "File Downloader  /"
						End If
					Else
						Me.Text = "File Downloader"
					End If
				End If
			End If
		End Sub
		#End Region

		#Region "CancelationIndicator"
		Private _cancelCancelationIndicator As Boolean = True

		Private Sub StopCancelationIndicator()
			_cancelCancelationIndicator = True
		End Sub

		Private Sub StartCancelationIndicator()
			_cancelCancelationIndicator = False
			ThreadPool.QueueUserWorkItem(Sub(o) AnimateCancelationIndicator(o))
		End Sub

		Private Sub AnimateCancelationIndicator(ByVal state As Object)
			Dim toggle? As Boolean = True
			Do While Not _cancelCancelationIndicator
				UpdateCancelationIndicator(toggle)
				Thread.Sleep(500)
				toggle = Not toggle
			Loop
			UpdateCancelationIndicator(Nothing)
		End Sub

		Private Sub UpdateCancelationIndicator(ByVal toggle? As Boolean)
			If Me.IsHandleCreated AndAlso (Not Me.IsDisposed) Then
				If Me.InvokeRequired Then
					Me.BeginInvoke(CType(Sub() UpdateCancelationIndicator(toggle), Action))
				Else
					If toggle.HasValue Then
						If toggle.Value Then
							_progress.Text = "Canceling..."
							Me.statusStrip1.Refresh()
						Else
							_progress.Text = ""
							Me.statusStrip1.Refresh()
						End If
					Else
						_progress.Text = ""
						Me.statusStrip1.Refresh()
					End If
				End If
			End If
		End Sub
		#End Region

	End Class
End Namespace
